import queue
import json
from flask import Flask
from flask_sqlalchemy import SQLAlchemy

app = Flask(__name__)
app.config["SQLALCHEMY_DATABASE_URI"] = "sqlite:///store.sqlite"
db = SQLAlchemy(app)


class Session(db.Model):
    __tablename__ = "sessions"
    id = db.Column(db.Integer, primary_key=True, index=True)
    name = db.Column(db.String(), unique=True)
    created_at = db.Column(db.Integer())
    events = db.relationship("Event", backref="session", passive_deletes=True)


class Event(db.Model):
    __tablename__ = "events"
    id = db.Column(db.Integer, primary_key=True, index=True)
    title = db.Column(db.String())
    message = db.Column(db.String())
    type = db.Column(db.String())
    session_id = db.Column(db.Integer, db.ForeignKey("sessions.id", ondelete="CASCADE"))


class SSE:
    def __init__(self):
        self.listeners = []

    def listen(self):
        self.listeners.append(queue.Queue(maxsize=10))
        return self.listeners[-1]

    def publish(self, message: dict, type: str):
        msg = f"data: {json.dumps(message)}\n\n"
        payload = f"event: {type}\n{msg}"

        for i in reversed(range(len(self.listeners))):
            try:
                self.listeners[i].put_nowait(payload)
            except queue.Full:
                del self.listeners[i]


sse = SSE()

with app.app_context():
    db.create_all()



import logging
from typing import Literal

import requests
from nanoid import generate

main_logger = logging.getLogger(__name__)

def logger():
    session = f"session-{generate(size=6)}"

    def log(
        title: str = "",
        message: str = "",
        type: Literal["info", "prompt", "completion", "warning", "error"] = "info",
    ) -> None:
        try:
            requests.post(
                "http://localhost:7000/event",
                json={"session": session, "title": title, "message": message, "type": type},
            )
        except requests.exceptions.ConnectionError:
            main_logger.info("Hayloft didn't start")

    return log



from peewee import SqliteDatabase, CharField, IntegerField, Model, ForeignKeyField
from pathlib import Path

path = str(Path(__file__).parent.resolve()) 
db = SqliteDatabase(f'{path}/store.db')

class Session(Model):
    name = CharField(unique=True)
    created_at = IntegerField()

    class Meta:
        database = db
        table_name = 'sessions'

class Event(Model):
    session = ForeignKeyField(Session, backref='events')
    title = CharField()
    message = CharField()
    type = CharField()

    class Meta:
        database = db
        table_name = 'events'




import queue
import json

class SSE:
    def __init__(self):
        self.listeners = []

    def listen(self):
        self.listeners.append(queue.Queue(maxsize=10))
        return self.listeners[-1]

    def publish(self, message: dict, type: str):
        msg = f"data: {json.dumps(message)}\n\n"
        payload = f"event: {type}\n{msg}"

        for i in reversed(range(len(self.listeners))):
            try:
                self.listeners[i].put_nowait(payload)
            except queue.Full:
                del self.listeners[i]


sse = SSE()





from gevent import monkey; monkey.patch_all()
from gevent import sleep
from bottle import app, static_file, request, response, GeventServer
from hayloft.schema import db, Session, Event
from hayloft.cors import EnableCors
from hayloft.sse import sse
from importlib.metadata import version
from pathlib import Path
from typing import Dict
import argparse
import time

app = app()
app.install(EnableCors()) 
path = str(Path(__file__).parent.resolve()) 

@app.get("/")
def index():
    return static_file("index.html", root=f"{path}/public")

@app.get("/assets/<file:path>")
def serve_assets(file):
    return static_file(file, root=f"{path}/public/assets")

@app.post("/event")
def create_event():
    body = request.json
    session_name = body.get("session")
    title = body.get("title")
    message = body.get("message")
    type = body.get("type")
    event: Event | None = None
    new_session: Session | None = None

    try:
        session = Session.select().where(Session.name == session_name).get()
        event = Event.create(session=session, title=title, message=message, type=type)
    except:
        created_at = int(time.time() * 1000)
        new_session = Session.create(name=session_name, created_at=created_at)
        event = Event.create(session=new_session, title=title, message=message, type=type)
    msg: Dict
    if new_session is not None:
        msg = {
            "session": {
                "id": new_session.id,
                "name": new_session.name,
                "created_at": new_session.created_at,
            },
            "event": None,
        }
    else:
        msg = {
            "event": {
                "id": event.id,
                "title": event.title,
                "message": event.message,
                "type": event.type,
                "session_id": event.session.id,
            },
            "session": None,
        }
    sse.publish(msg, type="stream")
    return "OK"

@app.get('/sessions')
def get_sessions():
    sessions = Session.select()
    return {"sessions": [{"id": s.id, "name": s.name, "created_at": s.created_at} for s in sessions]} 

@app.get("/sessions/<session_id:int>/events")
def get_events(session_id):
    events = Event.select().where(Event.session_id == session_id)
    return { 
        "events": [
            {"id": e.id, "title": e.title, "message": e.message, "type": e.type}
            for e in events
        ]
    }

@app.get("/listen")
def listen():
    response.set_header("Content-Type", "text/event-stream")
    response.set_header("Cache-Control", "no-cache")
    yield 'retry: 100\n\n' 
    messages = sse.listen()
    
    while True:
        msg = messages.get()
        yield msg
        sleep(0.05)

def start():
    db.connect()
    db.create_tables([Session, Event])

    print(f'\033[96mHayloft {version(__package__)} starting up, open in your browser http://localhost:7000\033[0m')
    print("Hit Ctrl-C to quit.")
    app.run(host='localhost', port=7000, server=GeventServer, quiet=True)

def cli():
    parser = argparse.ArgumentParser(description="Hayloft - UI tool for LLM frameworks")
    parser.add_argument("command", type=str, help="command to run", choices=["start"])
    parser.add_argument("-v", "--version", action="version", version=version(__package__))
    args = parser.parse_args()

    if args.command == "start":
        start()

if __name__ == '__main__':
    start()



import logging
from hayloft.logger import logger

class HayloftLogger(logging.Handler):
    def __init__(self):
        super().__init__()
        self.log = logger()
    
    def emit(self, record):
        type = "info"
        if record.levelno == logging.ERROR or record.levelno == logging.CRITICAL:
            type = "error"
        elif record.levelno == logging.WARNING:
            type = "warning"
        elif record.name.startswith("llama_index.llm_predictor"):
            type = "completion"
        elif record.name.startswith("llama_index.indices.response"):
            type = "completion" if record.msg.find(" response: ") != -1 else "prompt"

        self.log(title=record.name, message=record.msg, type=type)

def load_hayloft():
    logging.getLogger("llama_index").setLevel(logging.DEBUG)
    logging.getLogger().addHandler(HayloftLogger())



from flask import request, jsonify, send_file, send_from_directory, Response
from flask_cors import CORS
from hayloft.schema import app, db, sse, Event, Session
from importlib.metadata import version
import argparse
import time

CORS(app)  # for development

@app.route("/", methods=["GET"])
def index():
    return send_file("public/index.html")


@app.route("/assets/<path:path>", methods=["GET"])
def serve_assets(path):
    return send_from_directory("public/assets", path)


@app.route("/event", methods=["POST"])
def create_event():
    body = request.get_json()
    title = body["title"]
    message = body["message"]
    type = body["type"]
    event: Event | None = None
    sess: Session | None = None

    try:
        session = db.session.execute(
            db.select(Session).filter_by(name=body["session"])
        ).scalar_one()
        event = Event(title=title, message=message, type=type, session_id=session.id)
        db.session.add(event)
    except:
        created_at = int(time.time() * 1000)
        event = Event(title=title, message=message, type=type)
        sess = Session(name=body["session"], created_at=created_at, events=[event])
        db.session.add(sess)
    db.session.commit()
    msg: dict
    if sess is not None:
        msg = {
            "session": {
                "id": sess.id,
                "name": sess.name,
                "created_at": sess.created_at,
            },
            "event": None,
        }
    else:
        msg = {
            "event": {
                "id": event.id,
                "title": event.title,
                "message": event.message,
                "type": event.type,
                "session_id": event.session_id,
            },
            "session": None,
        }
    sse.publish(msg, type="stream")
    return "OK"


@app.route("/sessions", methods=["GET"])
def get_sessions():
    sessions = db.session.execute(db.select(Session)).scalars()
    return jsonify(
        [{"id": s.id, "name": s.name, "created_at": s.created_at} for s in sessions]
    )


@app.route("/sessions", methods=["DELETE"])
def remove_sessions():
    # todo it via cascade later
    db.session.query(Session).delete()
    db.session.query(Event).delete()
    db.session.commit()
    return ""


@app.route("/sessions/<int:session_id>/events", methods=["GET"])
def get_events(session_id):
    events = db.session.execute(
        db.select(Event).filter_by(session_id=session_id)
    ).scalars()
    return jsonify(
        [
            {"id": e.id, "title": e.title, "message": e.message, "type": e.type}
            for e in events
        ]
    )


@app.route("/listen", methods=["GET"])
def listen():
    def stream():
        messages = sse.listen()
        while True:
            msg = messages.get()
            yield msg

    return Response(stream(), mimetype="text/event-stream")


def cli():
    parser = argparse.ArgumentParser(description="Hayloft - UI tool for LLM frameworks")
    parser.add_argument("command", type=str, help="command to run", choices=["start"])
    parser.add_argument("-v", "--version", action="version", version=version(__package__))
    args = parser.parse_args()

    if args.command == "start":
        app.run(host="localhost", port=7000, debug=False)
